# Define a translation table to convert Vietnamese characters to their ASCII equivalents
VN_EN_TRANS = str.maketrans(
    "ÁÀẢÃẠĂẮẰẲẴẶÂẤẦẨẪẬĐÈÉẺẼẸÊẾỀỂỄỆÍÌỈĨỊÓÒỎÕỌÔỐỒỔỖỘƠỚỜỞỠỢÚÙỦŨỤƯỨỪỬỮỰÝỲỶỸỴ"
    "áàảãạăắằẳẵặâấầẩẫậđèéẻẽẹêếềểễệíìỉĩịóòỏõọôốồổỗộơớờởỡợúùủũụưứừửữựýỳỷỹỵ",
    "A"*17 + "D" + "E"*11 + "I"*5 + "O"*17 + "U"*11 + "Y"*5 +
    "a"*17 + "d" + "e"*11 + "i"*5 + "o"*17 + "u"*11 + "y"*5,
    chr(774) + chr(770) + chr(795) + chr(769) + chr(768) + chr(777) + chr(771) + chr(803)
)

UPPER_ALPHABET = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
LOWER_ALPHABET = "abcdefghijklmnopqrstuvwxyz"


def remove_vn_diacritics(txt: str) -> str:
    """
    Converts Vietnamese characters with diacritics to their corresponding
    ASCII equivalents.

    Args:
        txt (str): The Vietnamese text to be converted.

    Returns:
        str: The converted text with diacritics removed.
    """
    return txt.translate(VN_EN_TRANS)



from typing import Literal
import numpy as np


def psi(expected: np.ndarray, actual: np.ndarray, bucket_type: Literal["bins", "quantiles"] = "quantiles", n_bins: int = 10) -> float:
    """
    Calculate the Population Stability Index (PSI) for two arrays.

    Parameters
    ----------
    expected : np.ndarray
        Array of expected values.
    actual : np.ndarray
        Array of actual values.
    bucket_type : Literal["bins", "quantiles"], optional
        Binning strategy. Accepts two options: 'bins' and 'quantiles'. Defaults to 'quantiles'.
        'bins': input arrays are split into bins with equal and fixed steps based on 'expected' array.
        'quantiles': input arrays are binned according to 'expected' array with given number of n_bins.
    n_bins : int, optional
        Number of buckets for binning. Defaults to 10.

    Returns
    -------
    float
        A single float number representing the PSI value.
    """
    if bucket_type == "bins":
        min_val = expected.min()
        max_val = expected.max()
        bins = np.linspace(min_val, max_val, n_bins + 1)
    elif bucket_type == "quantiles":
        percentage = np.arange(0, n_bins + 1) / n_bins * 100
        bins = np.percentile(expected, percentage)

    # Calculate frequencies
    expected_percents = np.histogram(expected, bins)[0] / len(expected)
    actual_percents = np.histogram(actual, bins)[0] / len(actual)

    return psi_from_distribution(expected_percents, actual_percents)


def psi_from_distribution(expected_percents: np.ndarray, actual_percents: np.ndarray) -> float:
    """
    Calculate PSI from distributions of expected and actual percentages.

    Parameters
    ----------
    expected_percents : np.ndarray
        Array of expected percentages.
    actual_percents : np.ndarray
        Array of actual percentages.

    Returns
    -------
    float
        A single float number representing the PSI value.
    """
    # Normalize to sum to 1
    expected_percents = expected_percents / np.sum(expected_percents)
    actual_percents = actual_percents / np.sum(actual_percents)

    # Clip frequencies to avoid division by zero
    expected_percents = np.clip(expected_percents, a_min=0.00001, a_max=None)
    actual_percents = np.clip(actual_percents, a_min=0.00001, a_max=None)

    # Calculate PSI
    psi_value = (expected_percents - actual_percents) * np.log(expected_percents / actual_percents)
    psi_value = np.sum(psi_value)

    return psi_value



from typing import List

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, RegressorMixin


class VotingModel(BaseEstimator, RegressorMixin):
    """
    An ensemble learning model that combines multiple base estimators to make predictions.
    Implements both the BaseEstimator and RegressorMixin from scikit-learn.

    Parameters
    ----------
    estimators : List[BaseEstimator]
        A list of scikit-learn estimators that will be used to make predictions.
        Each estimator in the list should implement the `predict` method, and if `predict_proba` 
        is to be used, they should also implement the `predict_proba` method.
    """

    def __init__(self, estimators: List[BaseEstimator]):
        """
        Initialize the VotingModel with a list of base estimators.

        Parameters
        ----------
        estimators : List[BaseEstimator]
            A list of scikit-learn estimators.
        """
        super().__init__()
        self.estimators = estimators

    def fit(self, X, y=None):
        """
        Fit the model. This method does not perform any fitting and is included to 
        maintain compatibility with scikit-learn's interface.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            The input samples.

        y : array-like, shape (n_samples,), optional (default=None)
            The target values.

        Returns
        -------
        self : object
            Returns self.
        """
        return self

    def predict(self, X: pd.DataFrame):
        """
        Predict using the model by averaging the predictions of all base estimators.

        Parameters
        ----------
        X : pd.DataFrame
            The input samples.

        Returns
        -------
        y_pred : np.ndarray
            The averaged predictions.
        """
        y_preds = [estimator.predict(X) for estimator in self.estimators]
        return np.mean(y_preds, axis=0)

    def predict_proba(self, X: pd.DataFrame):
        """
        Predict class probabilities using the model by averaging the probability 
        predictions of all base estimators. This method is applicable only if all 
        base estimators implement `predict_proba`.

        Parameters
        ----------
        X : pd.DataFrame
            The input samples.

        Returns
        -------
        y_pred_proba : np.ndarray
            The averaged probability predictions.
        """
        y_preds = [estimator.predict_proba(X) for estimator in self.estimators]
        return np.mean(y_preds, axis=0)



from typing import List
import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.feature_selection import f_classif


class FeatureRemover(TransformerMixin, BaseEstimator):
    """
    Transformer that removes specified columns from a DataFrame.

    Parameters
    ----------
    droped_cols : List[str], optional (default=[])
        List of column names to be removed.
    """

    def __init__(self, droped_cols: List[str] = []):
        """
        Initialize the FeatureRemover with the specified columns to be removed.

        Parameters
        ----------
        droped_cols : List[str], optional (default=[])
            List of column names to be removed.
        """
        self.droped_cols = droped_cols

    def fit(self, X, y=None):
        """
        Fit the transformer. This method does not perform any fitting and is included 
        to maintain compatibility with scikit-learn's interface.

        Parameters
        ----------
        X : pd.DataFrame
            The input samples.

        y : array-like, shape (n_samples,), optional (default=None)
            The target values.

        Returns
        -------
        self : object
            Returns self.
        """
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Transform the input DataFrame by removing the specified columns.

        Parameters
        ----------
        X : pd.DataFrame
            The input samples.

        Returns
        -------
        X : pd.DataFrame
            The transformed DataFrame with specified columns removed.
        """
        columns = list(set(self.droped_cols).intersection(X.columns))
        return X.drop(columns=columns)


class ColinearFeatureRemover(FeatureRemover):
    """
    Transformer that removes collinear features from a DataFrame based on a correlation threshold.

    Parameters
    ----------
    correlation_threshold : float, optional (default=0.9)
        The correlation threshold above which features are considered collinear and are removed.
    """

    def __init__(self, correlation_threshold=0.9):
        """
        Initialize the ColinearFeatureRemover with the specified correlation threshold.

        Parameters
        ----------
        correlation_threshold : float, optional (default=0.9)
            The correlation threshold above which features are considered collinear and are removed.
        """
        super().__init__([])
        self.correlation_threshold = correlation_threshold

    def fit(self, X: pd.DataFrame, y: pd.Series = None):
        """
        Fit the transformer by identifying collinear features to be removed.

        Parameters
        ----------
        X : pd.DataFrame
            The input samples.

        y : pd.Series, optional (default=None)
            The target values.

        Returns
        -------
        self : object
            Returns self.
        """
        correlations = []
        for col in X.columns:
            correlations.append(np.abs(y.corr(X[col])))
        df_clusters = pd.DataFrame(
            zip(X.columns, correlations),
            columns=['feature', 'correlation']
        )
        df_clusters = df_clusters\
            .sort_values(by='correlation', ascending=False)\
            .reset_index(drop=True)
        df_clusters = df_clusters[~df_clusters["correlation"].isna()]
        to_remove_list = []
        corr = X[df_clusters['feature']].corr()

        for idx, col_a in enumerate(corr.columns):
            if col_a not in to_remove_list:
                for col_b in corr.columns[idx+1:]:
                    if corr[col_a][col_b] > self.correlation_threshold:
                        to_remove_list.append(col_b)

        self.droped_cols = to_remove_list
        return self


class UnivariateFeatureRemover(FeatureRemover):
    def __init__(self, score_func: callable = f_classif, threshold=0.05):
        super().__init__([])
        self.score_func = score_func
        self.threshold = threshold
        self.feature_importance = None

    def fit(self, X: pd.DataFrame, y: pd.Series = None):
        f_statistic, p_values = self.score_func(X, y)
        self.feature_importance = pd.DataFrame({
            "features": X.columns,
            "f_statistic": f_statistic,
            "p_values": p_values
        }).fillna({
            "f_statistic": 0,
            "p_values": 1,
        })
        self.droped_cols = list(self.feature_importance[self.feature_importance["p_values"] <= self.threshold]["features"])
        return self



from typing import Union


def round(val: Union[int, float], type: type = None):
    if type == int or isinstance(val, int):
        num_digits = len(str(int(val)))
        round_to = (num_digits - 2) * -1
        if num_digits > 2:
            return round(val, round_to)
        else:
            return int(val)
    elif type == float or isinstance(val, int):
        return round(val, 2)
    else:
        pass
    return val



from typing import List, Set, Dict, Tuple, Union
import pandas as pd
import numpy as np
from datetime import datetime
from google.cloud import bigquery
from caketool.metric import psi_from_distribution
from caketool.utils import str_utils, num_utils


class FeatureMonitor:
    """
    A class to monitor and analyze features in a dataset.
    """

    def __init__(self, project, location, dataset="mlops_motinor") -> None:
        """
        Initialize the FeatureMonitor class.

        Parameters
        ----------
        project : str
            Google Cloud project name.
        location : str
            Location of the Google Cloud resources.
        dataset : str, optional
            Dataset name in BigQuery. Defaults to 'mlops_motinor'.
        """
        self.project = project
        self.location = location
        self.dataset = dataset
        self.MISSING = 'cake.miss'
        self.OTHER = 'cake.other'
        self.bq_client = bigquery.Client(project=self.project, location=self.location)

    def normalize_data(
        self, df: pd.DataFrame, inplace: bool = False,
        cate_missing_values: Set[str] = {'-1', '-100', 'unknown', ''}
    ) -> pd.DataFrame:
        """
        Normalize the data by filling missing values and standardizing feature types.

        Parameters
        ----------
        df : pd.DataFrame
            DataFrame to be normalized.
        inplace : bool, optional
            Whether to modify the DataFrame in place. Defaults to False.
        cate_missing_values : Set[str], optional
            Set of categorical missing values. Defaults to {'-1', '-100', 'unknown', ''}.

        Returns
        -------
        pd.DataFrame
            Normalized DataFrame.
        """
        if not inplace:
            df = df.copy()
        # Fill missing value
        df = df.fillna(-100)
        # Norm categorical feature type
        categorical_features: List[str] = []
        for col in df.columns:
            try:
                df[col] = pd.to_numeric(df[col])
            except (ValueError, KeyError):
                categorical_features.append(col)
        df[categorical_features] = df[categorical_features].apply(lambda x: x.astype(str).str.lower())

        # Find numerical features
        numerical_features: List[str] = df.select_dtypes([int, float]).columns
        # Handle infinity value
        for col in list(set(df[numerical_features].columns.to_series()[np.isinf(df[numerical_features]).any()])):
            df[col] = df[col].apply(lambda x: -100 if x == np.inf else x)
        # Norm float feature
        float_features: List[str] = []
        for col in numerical_features:
            if df[col].fillna(0).nunique() > round(df[col].fillna(0)).nunique():
                float_features.append(col)
        df[float_features] = df[float_features].astype(float)
        # Norm int feature
        int_features = list(set(df.columns) - set(categorical_features) - set(float_features))
        df[int_features] = df[int_features].astype(int)
        # Handle numeric columns
        for col in numerical_features:
            df[col] = df[col].apply(lambda x: -100 if x < 0 else x)
        # Handle categorical columns
        for col in categorical_features:
            df[col] = df[col].apply(
                lambda x: self.MISSING
                if x in cate_missing_values
                else str_utils.remove_vn_diacritics(x).lower().strip()
            )

        return df

    def create_bin_data(
        self,
        df: pd.DataFrame,
        n_bins=10
    ) -> Dict[str, List[Union[float, str]]]:
        """
        Create bin data for numerical and categorical features.

        Parameters
        ----------
        df : pd.DataFrame
            DataFrame to create bin data from.
        n_bins : int, optional
            Number of bins. Defaults to 10.

        Returns
        -------
        Dict[str, List[Union[float, str]]
            Dictionary of bin thresholds for each feature.
        """
        int_features = set(df.select_dtypes([int]).columns)
        float_features = set(df.select_dtypes([float]).columns)
        categorical_features = list(df.select_dtypes([object]).columns)
        bin_thresholds = dict()
        # Bin num features
        for f in [*int_features, *float_features]:
            series = df[f]
            series = series[(series > 0) & (~series.isna())]
            if len(series) == 0:
                bin_thresholds[f] = []
                continue
            percentage = np.linspace(0, 100, n_bins + 1)
            bins = np.percentile(series, percentage)
            if f in int_features:
                bins = [num_utils.round(e, int) for e in bins]
            if f in float_features:
                bins = [num_utils.round(e, float) for e in bins]

            if bins[0] != 0:
                bins = [0, *bins]
            bins = np.unique(bins)
            if len(bins) >= 2:
                bins[-1] = bins[-1] + 1e-10
            bin_thresholds[f] = bins
        # Bin cate features
        for f in categorical_features:
            series = df[f]
            bins = series.value_counts()[:n_bins].index.to_list()
            bins = sorted(set([self.MISSING, self.OTHER, *bins]))
            bin_thresholds[f] = bins

        return bin_thresholds

    def calculate_distribution(
        self,
        df: pd.DataFrame,
        bin_thresholds: Dict[str, np.ndarray]
    ) -> pd.DataFrame:
        """
        Calculate the distribution of features based on bin thresholds.

        Parameters
        ----------
        df : pd.DataFrame
            DataFrame to calculate distribution from.
        bin_thresholds : Dict[str, np.ndarray]
            Dictionary of bin thresholds for each feature.

        Returns
        -------
        pd.DataFrame
            DataFrame of feature distributions.
        """
        hists = []
        categorical_features = [k for k, v in bin_thresholds.items() if len(v) > 0 and isinstance(v[0], str)]
        numerical_features = [k for k, v in bin_thresholds.items() if len(v) > 0 and not isinstance(v[0], str)]

        for f in numerical_features:
            series = df[f]
            hist, _ = np.histogram(series, [-np.inf, *bin_thresholds[f], np.inf])
            hists.append([f, "num", list(map(str, bin_thresholds[f])), hist.astype(np.int64).tolist()])
        for f in categorical_features:
            series = df[f]
            bins = bin_thresholds[f]
            series = series.apply(lambda x: x if x in bins else self.OTHER)
            vc = series.value_counts()
            for bin_name in bins:
                if bin_name not in vc.index:
                    vc.loc[bin_name] = 0
            vc = vc.reindex(bins)
            hists.append([f, "cate", vc.index.to_list(), vc.tolist()])

        return pd.DataFrame(hists, columns=["feature_name", "type", "bins", "histogram"])

    def store_distribution(
            self,
            df_distribution: pd.DataFrame,
            dataset_id: str,
            bq_table_name="feature_distribution"
    ) -> None:
        """
        Store the feature distribution in BigQuery.

        Parameters
        ----------
        df_distribution : pd.DataFrame
            DataFrame of feature distributions.
        dataset_id : str
            Dataset ID.
        bq_table_name : str, optional
            Table name in BigQuery. Defaults to 'feature_distribution'.
        """
        df_distribution = df_distribution.copy()
        df_distribution["dataset_id"] = dataset_id
        job_config = bigquery.LoadJobConfig(
            schema=[
                bigquery.SchemaField("dataset_id", 'STRING', 'REQUIRED'),
                bigquery.SchemaField("feature_name", 'STRING', 'REQUIRED'),
                bigquery.SchemaField("type", 'STRING', 'REQUIRED'),
                bigquery.SchemaField("bins", 'STRING', 'REPEATED'),
                bigquery.SchemaField("histogram", 'INTEGER', 'REPEATED'),
            ],
            clustering_fields=["dataset_id"]
        )
        table_id = f"{self.project}.{self.dataset}.{bq_table_name}"
        self.bq_client.query(f"DELETE FROM {table_id} WHERE dataset_id = '{dataset_id}'").result()
        self.bq_client.load_table_from_dataframe(df_distribution, table_id, job_config=job_config)

    def load_distribution(
        self, dataset_id: str, table_name: str = "feature_distribution"
    ) -> Tuple[pd.DataFrame, Dict[str, List[Union[str, float]]]]:
        """
        Load the feature distribution from BigQuery.

        Parameters
        ----------
        dataset_id : str
            Dataset ID.
        table_name : str, optional
            Table name in BigQuery. Defaults to 'feature_distribution'.

        Returns
        -------
        pd.DataFrame
            DataFrame of feature distributions.
        """
        table_id = f"{self.project}.{self.dataset}.{table_name}"
        df_distribution: pd.DataFrame = self.bq_client.query(f"""
            SELECT * FROM {table_id}
            WHERE dataset_id = '{dataset_id}'
        """).to_dataframe()
        bin_thresholds = {}
        for _, row in df_distribution.iterrows():
            if row["type"] == "num":
                bin_thresholds[row["feature_name"]] = list(map(float, row["bins"]))
            if row["type"] == "cate":
                bin_thresholds[row["feature_name"]] = list(map(str, row["bins"]))

        return df_distribution, bin_thresholds

    def calculate_feature_drift(self, dev_distribution: pd.DataFrame, test_distribution: pd.DataFrame) -> pd.DataFrame:
        """
        Calculate the feature drift between two distributions.

        Parameters
        ----------
        dev_distribution : pd.DataFrame
            DataFrame of the development distribution.
        test_distribution : pd.DataFrame
            DataFrame of the test distribution.

        Returns
        -------
        pd.DataFrame
            DataFrame of feature drift with PSI values.
        """
        comp_psi = pd.merge(
            dev_distribution, test_distribution,
            how="inner",
            on=["feature_name", "type"],
            suffixes=["_dev", "_test"]
        )

        for col in ["histogram_dev", "histogram_test"]:
            if comp_psi[col].apply(lambda s: s.sum()).nunique() != 1:
                raise ValueError(f"Sizing of {col} is not consistent between the features of the dataset!")

        wrong_bins_features = comp_psi[comp_psi.apply(lambda s: not (s["bins_dev"] == s["bins_test"]).all(), axis=1)]["feature_name"]
        if len(wrong_bins_features) > 0:
            raise ValueError(f"Bins is not the same between 2 datasets. {','.join(wrong_bins_features)}")

        comp_psi["psi"] = comp_psi.apply(lambda r: psi_from_distribution(r["histogram_dev"], r["histogram_test"]), axis=1)
        return comp_psi.sort_values(by="psi", ascending=False)

    def transform_distribution_to_looker(
        self,
        score_type: str,
        dataset_type: str,
        version_type: str,
        version: str,
        df_distribution,
        bq_table_name="feature_distribution_looker"
    ) -> pd.DataFrame:
        """
        Store the feature distribution in Looker format in BigQuery.

        Parameters
        ----------
        score_type : str
            Type of score.
        dataset_type : str
            Type of dataset.
        version_type : str
            Type of version.
        version : str
            Version identifier.
        df_distribution : pd.DataFrame
            DataFrame of feature distributions.
        bq_table_name : str, optional
            Table name in BigQuery. Defaults to 'feature_distribution_looker'.

        Returns
        -------
        pd.DataFrame
            DataFrame in Looker format.
        """
        ls_row = []
        for _, row in df_distribution.iterrows():
            feature_name = row["feature_name"]
            bins = row["bins"]
            histogram = row["histogram"]
            total = np.sum(histogram)
            if row["type"] == "cate":
                segment_label = [". ".join(e) for e in zip(str_utils.UPPER_ALPHABET, bins)]
                dist = zip(segment_label, histogram)
            if row["type"] == "num":
                bins = [round(float(e), 2) for e in row["bins"]]
                bins = [int(e) if e.is_integer() else e for e in bins]
                if len(bins) <= 0:
                    continue
                segment_label = []
                for s, e in zip(bins[:-1], bins[1:]):
                    segment_label.append(f"[{s}, {e})")
                if len(segment_label) > 0:
                    segment_label[-1] = segment_label[-1].replace(")", "]")
                segment_label = [f"A. missing", *[". ".join(e) for e in zip(str_utils.UPPER_ALPHABET[2:], segment_label)], f"B. other"]
                dist = zip(segment_label, histogram)
            for k, v in dist:
                ls_row.append([
                    feature_name, k, v, total, v/total
                ])

        df_looker = pd.DataFrame(ls_row, columns=["feature_name", "segment", "count", "total", "percent"])
        df_looker["score_type"] = score_type
        df_looker["dataset_type"] = dataset_type
        df_looker["version_type"] = version_type
        df_looker["version"] = version
        df_looker["utc_update_at"] = datetime.now()

        job_config = bigquery.LoadJobConfig(
            schema=[
                bigquery.SchemaField("score_type", 'STRING', 'REQUIRED'),
                bigquery.SchemaField("dataset_type", 'STRING', 'REQUIRED'),
                bigquery.SchemaField("version_type", 'STRING', 'REQUIRED'),
                bigquery.SchemaField("version", 'STRING', 'REQUIRED'),
                bigquery.SchemaField("feature_name", 'STRING', 'REQUIRED'),
                bigquery.SchemaField("segment", 'STRING', 'REQUIRED'),
                bigquery.SchemaField("count", 'INTEGER', 'REQUIRED'),
                bigquery.SchemaField("total", 'INTEGER', 'REQUIRED'),
                bigquery.SchemaField("percent", 'FLOAT', 'REQUIRED'),
                bigquery.SchemaField("utc_update_at", 'DATETIME', 'REQUIRED'),
            ],
            clustering_fields=["score_type", "dataset_type", "version_type", "version"]
        )
        table_id = f"{self.project}.{self.dataset}.{bq_table_name}"
        self.bq_client.query(f"""
            DELETE FROM {table_id}
            WHERE score_type = '{score_type}'
            AND dataset_type = '{dataset_type}'
            AND version_type = '{version_type}'
            AND version = '{version}'
        """).result()
        self.bq_client.load_table_from_dataframe(df_looker, table_id, job_config=job_config)
        return df_looker



import pandas as pd
import numpy as np
from datetime import datetime
from google.cloud import bigquery


class ModelMonitor:

    def __init__(self, project: str, location: str, dataset: str = "mlops_motinor") -> None:
        self.project = project
        self.location = location
        self.dataset = dataset
        self.bq_client = bigquery.Client(project=self.project, location=self.location)

    def calc_score_distribution(
            self,
            score: np.ndarray,
            bins=[
                0, 0.04, 0.05, 0.06, 0.07, 0.08, 0.09, 0.1, 0.125,
                0.15, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1
            ]):

        total = len(score)
        histogram = np.histogram(score, bins)[0]
        percent = histogram / total
        return pd.DataFrame({
            "segment": list(map(str, bins[1:])),
            "count": histogram,
            "total": [total] * len(histogram),
            "percent": percent,
        })

    def store_distribution_looker(
        self,
        score_type: str,
        dataset_type: str,
        version_type: str,
        version: str,
        df_distribution: pd.DataFrame,
        bq_table_name="score_distribution_looker"
    ) -> pd.DataFrame:
        df_looker = df_distribution.copy()
        df_looker["score_type"] = score_type
        df_looker["dataset_type"] = dataset_type
        df_looker["version_type"] = version_type
        df_looker["version"] = version
        df_looker["utc_update_at"] = datetime.now()

        job_config = bigquery.LoadJobConfig(
            schema=[
                bigquery.SchemaField("score_type", 'STRING', 'REQUIRED'),
                bigquery.SchemaField("dataset_type", 'STRING', 'REQUIRED'),
                bigquery.SchemaField("version_type", 'STRING', 'REQUIRED'),
                bigquery.SchemaField("version", 'STRING', 'REQUIRED'),
                bigquery.SchemaField("segment", 'STRING', 'REQUIRED'),
                bigquery.SchemaField("count", 'INTEGER', 'REQUIRED'),
                bigquery.SchemaField("total", 'INTEGER', 'REQUIRED'),
                bigquery.SchemaField("percent", 'FLOAT', 'REQUIRED'),
                bigquery.SchemaField("utc_update_at", 'DATETIME', 'REQUIRED'),
            ],
            clustering_fields=["score_type", "dataset_type", "version_type", "version"]
        )
        table_id = f"{self.project}.{self.dataset}.{bq_table_name}"
        self.bq_client.query(f"""
            DELETE FROM {table_id}
            WHERE score_type = '{score_type}'
            AND dataset_type = '{dataset_type}'
            AND version_type = '{version_type}'
            AND version = '{version}'
        """)
        self.bq_client.load_table_from_dataframe(df_looker, table_id, job_config=job_config)
        return df_looker
